# setup.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)
"""pgn-read setup file.

Retained so 'python setup.py ...' can be used by my build scripts until
conversion to 'python -m pip ...'.

All setup() arguments are set in setup.cfg file.
"""

from setuptools import setup

if __name__ == "__main__":

    setup()
